"""
@file
@brief Various conversion functions.
"""
import pandas
from pyquickhelper.loghelper import noLOG
from ..sql.database_main import Database


def dBase2df(file, encoding="cp437"):
    """
    converts a dBase file into a list of dataframe (one per table)

    @param      file        file name
    @param      encoding    table encoding
    @return                 list of dataframes (pandas)

    The module relies on `dbfread <https://pypi.python.org/pypi/dbfread/>`_.
    """
    import dbfread  # pylint: disable=C0415
    table = dbfread.DBF(file, load=False, encoding=encoding)
    res = list(table)
    return pandas.DataFrame(res)


def dBase2sqllite(
        db, table, encoding="cp437", overwrite_table=None, fLOG=noLOG):
    """
    Put all rows from a dBase database into sqlite

    Add a dbase table to an open sqlite database.

    @param      db                  cursor on SQLite or file name
    @param      table               DBF object or filename
    @param      encoding            encoding if table is a filename
    @param      overwrite_table     overwrite the table name
    @param      fLOG                logging function, to see the progress

    The table will be removed if it exists.
    """

    typemap = {
        'F': 'FLOAT',
        'L': 'BOOLEAN',
        'I': 'INTEGER',
        'C': 'TEXT',
        'N': 'REAL',  # because it can be integer or float
        'M': 'TEXT',
        'D': 'DATE',
        'T': 'DATETIME',
        '0': 'INTEGER',
    }

    if isinstance(db, str):
        cursor = Database(db, LOG=fLOG)
        cursor.connect()
    else:
        cursor = db

    if isinstance(table, str):
        import dbfread  # pylint: disable=C0415
        table = dbfread.DBF(table, load=False, encoding=encoding)

    cursor.execute('drop table if exists %s' % table.name)

    field_types = {}
    for f in table.fields:
        field_types[f.name] = typemap.get(f.type, 'TEXT')

    table_name = overwrite_table if overwrite_table is not None else table.name

    # Create the table
    #
    defs = ', '.join(['%s %s' % (f, field_types[f])
                      for f in table.field_names])
    sql = 'create table %s (%s)' % (table_name, defs)
    cursor.execute(sql)

    # Create data rows
    refs = ', '.join([':' + f for f in table.field_names])
    sql = 'insert into %s values (%s)' % (table_name, refs)

    for i, rec in enumerate(table):
        cursor._connection.execute(sql, list(rec.values()))
        if i % 20000 == 0:
            fLOG("moving line ", i, " to table", table_name)

    if isinstance(db, str):
        cursor.commit()
        cursor.close()



from antlr4 import *


class CSharpParserBase(Parser):
    def __init__(self, input, output):
        Parser.__init__(self, input, output)

    def IsLocalVariableDeclaration(self):
        # as Test.CSharpParser.Local_variable_declarationContext
        local_var_decl = self.Context
        if local_var_decl is None:
            return True
        local_variable_type = local_var_decl.local_variable_type()
        if local_variable_type is None:
            return True
        if local_variable_type.GetText() == "var":
            return False
        return True



"""
@file
@brief Helpers around language grammar.
This module requires `antlr4 <https://pypi.python.org/pypi/antlr4-python3-runtime/>`_.
"""
from antlr4 import ParseTreeListener


class TreeStringListener(ParseTreeListener):

    """
    This class is an attempt to run through the tree and convert it into
    a string.
    """

    def __init__(self, parser):
        """
        constructor

        @param      parser      parser used to parse the code
        """
        ParseTreeListener.__init__(self)
        self.buffer = []
        self.level = 0
        self.parser = parser

    def visitTerminal(self, node):
        """
        event
        """
        text = ("    " * self.level) + "v " + str(node.symbol)
        self.buffer.append(text)

    def visitErrorNode(self, node):
        """
        event
        """
        text = ("    " * self.level) + "error: " + str(node)
        self.buffer.append(text)

    def enterEveryRule(self, ctx):
        """
        event
        """
        if "ruleIndex" in ctx.__dict__:
            text = ("    " * self.level) + "+ " + \
                self.parser.ruleNames[
                ctx.ruleIndex] + ", LT(1)=" + self.parser._input.LT(1).text
        else:
            text = ("    " * self.level) + "+ " + \
                ", LT(1)=" + self.parser._input.LT(1).text
        self.buffer.append(text)
        self.level += 1

    def exitEveryRule(self, ctx):
        """
        event
        """
        self.level -= 1
        if "ruleIndex" in ctx.__dict__:
            text = ("    " * self.level) + "- " + \
                self.parser.ruleNames[ctx.ruleIndex] + \
                ", LT(1)=" + self.parser._input.LT(1).text
        else:
            text = ("    " * self.level) + "- " + \
                ", LT(1)=" + self.parser._input.LT(1).text
        self.buffer.append(text)

    def __str__(self):
        """
        usual
        """
        return "\n".join(self.buffer)



"""
@file
@brief  helpers about python types
"""

import decimal


def guess_type_value(x, none=None):
    """
    Guesses the type of a value.

    @param      x           type
    @param      none        if True and all values are empty, return None
    @return                 type

    @warning if an integer starts with a zero, then it is a string.
    """
    if x is None:
        return none
    try:
        int(x)
        if isinstance(x, str) and x[0] == '0' and len(x) > 1:
            return str
        else:
            return int if isinstance(x, int) or len(x) < 9 else str
    except (ValueError, TypeError):
        try:
            x = float(x)
            return float
        except ValueError:
            if none:  # pragma: no cover
                if x is None:
                    return None
                try:
                    if len(x) > 0:
                        return str
                    else:
                        return None
                except Exception:
                    return None
            else:
                return str


def guess_type_value_type(none=True):
    """
    @param      none        if True and all values are empty, return None
    @return                 the list of types recognized by guess_type_value
    """
    return [None, str, int, float] if none else [str, int, float]


def get_default_value_type(ty, none=True):
    """
    @param      ty          type in guess_type_value_type
    @param      none        if True and all values are empty, return None
    @return                 a default value for this type
    """
    if ty is None and none:
        return None
    elif ty == str:
        return ""
    elif ty == int:
        return 0
    elif ty == decimal.Decimal:
        return decimal.Decimal(0)
    elif ty == float:
        return 0.0
    else:
        raise TypeError(  # pragma: no cover
            "type expected in " + str(guess_type_value_type()))



from antlr4 import *


class CSharpLexerBase(Lexer):
    def __init__(self, input, output):
        Lexer.__init__(self, input, output)

        self.interpolatedStringLevel = 0
        self.interpolatedVerbatiums = []
        self.curlyLevels = []
        self.verbatium = False

    def OnInterpolatedRegularStringStart(self):
        self.interpolatedStringLevel += 1
        self.interpolatedVerbatiums.append(False)
        verbatium = False

    def OnInterpolatedVerbatiumStringStart(self):
        self.interpolatedStringLevel += 1
        self.interpolatedVerbatiums.append(True)
        self.verbatium = True

    def OnOpenBrace(self):
        if self.interpolatedStringLevel > 0:
            self.curlyLevels.append(curlyLevels.pop() + 1)

    def OnCloseBrace(self):
        if self.interpolatedStringLevel > 0:
            self.curlyLevels.append(curlyLevels.pop() - 1)
            if self.curlyLevels.Peek() == 0:
                self.curlyLevels.pop()
                self.Skip()
                popMode()

    def OnColon(self):
        if self.interpolatedStringLevel > 0:
            ind = 1
            switchToFormatString = True
            while _input.La(ind) != '}':
                if _input.La(ind) == ':' or _input.La(ind) == ')':
                    switchToFormatString = False
                    break
                ind += 1
            if switchToFormatString:
                self.Mode(Test.CSharpLexer.INTERPOLATION_FORMAT)

    def OpenBraceInside(self):
        self.curlyLevels.append(1)

    def OnDoubleQuoteInside(self):
        self.interpolatedStringLevel -= 1
        self.interpolatedVerbatiums.pop()
        self.verbatium = self.interpolatedVerbatiums.Peek(
        ) if self.interpolatedVerbatiums.Count() > 0 else False

    def OnCloseBraceInside(self):
        self.curlyLevels.pop()

    def IsRegularCharInside(self):
        return not self.verbatium

    def IsVerbatiumDoubleQuoteInside(self):
        return self.verbatium
